// --- [ ir_ma_receive.c ] 18T -----------------------------------------------
//       tab = 3, 24.01.09
// Fr den Master genderte Routine zum Empfangen von 2 Byte via Manchester-Modulation auf 36 KHz
// nicht kompatibel mit RC5-Standard

// Benutzt den Timer0 zur Bestimmung der Impulsdauer und fr Zeitbegrenzung (Timer0_ovf)

// Arbeitet interruptgesteuert:
// Wartet auf die fallende Flanke eines Startbits, dann wird der Timer gestartet mit TCNT0 = 0
// es wird auf 1/2 Takt gewartet,

// Bei jedem folgenden Interrupt wird die zeitliche Distanz zum
// Vorgnger getestet:
// bei 1/2 Taktabstand liegt nur Flankenwechsel vor,
// bei 1/1 Taktabstand liegt der relevante Flankenwechsel in Taktmitte vor,
// bei fallender Flanke wurde eine '1', bei steigender Flanke eine '0' bertragen.

// Die eingehenden Bits werden von rechts nach links in eine Variable geschoben.

// Nach 16 Bits ist Ende erreicht.
// Wenn weniger Bits eintreffen, lst der Timer0_OVF nach 1.5 Takten einen Fehler aus,
// treffen mehr Bits ein, dann wird das Ergebnis berschrieben und lst dann
// wenn nicht zufllig wieder genau 16 Bit eintreffen - einen Fehler aus.

// Die Dauer eines Bits betrgt 18 Takte bei 36KHz low sowie
// 18 Takte high: (18 + 18) / 36000 = 36 / 36000 =  1000us
// bei Prescale des Timers von '64': 64/8.000.000 = 8us/Timerwert
// der TCNTO erhht sich innerhalb eines ganzen Taktes um 1000/8us = 125 

// Achtung Verstndishrde:
// wenn das zuletzt eingehende Bit eine '1' ist, dann folgt auf die fallende
// Flanke noch eine steigende (das Signal geht in den Ruhezustand HIGH) zurck.
// Diese Flanke lst einen Interrupt aus, der als Start fr ein folgendes Byte
// zhlt ... 
// Um dies Problem zu lsen, wird nach dem letzten signifikanten Levelwechsel
// der INT0 Sense Control von 'any logic change' auf 'falling edge' geschaltet,
// dadurch wird ein folgende steigende Flanke ignoriert.

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "ir_master.h"

// lokale Variabelen
volatile uint8_t IR_in;						// 8Bit Daten
volatile uint8_t IR_start;					// Flag fr den Status des Empfngers
volatile uint8_t IR_error;					// Fehlercode
volatile uint8_t IR_ok;						// zeigt an, dass gltige Daten vorliegen
volatile uint8_t bitcount;					// Zhler fr eingegangene Bits

// ----------------------------------------------------------------------------
// Den IR-Empfnger konfigurieren.
// Der Timer0 luft mit Prescaler 64 -> 64 / 4.000.000  -> 16us/Tick
void IR_init(void)
{
TCCR0A = 0;										// Einstellung fr PWM rausnehmen
TIMSK = (1<<TOIE0);							// Timer0 Overflow Interrupt Enable
PCMSK = (1<<PCINT4);							// Pin Change an PinB.4 aktivieren
GIMSK = (1<<PCIE);							// Pin Change Interrupt enablen
sei();
}

// ----------------------------------------------------------------------------
// den IR-Empfang deaktivieren
void IR_rx_off(void)
{
GIMSK = 0;
PCMSK = 0;
}

// ----------------------------------------------------------------------------
// exportiert, ob Daten empfangen wurden
uint8_t IR_complete(void)
{
return IR_ok;
}

// ----------------------------------------------------------------------------
// exportiert empfangenes Byte und lscht Flag IR_ok, 
// damit neue Daten erkannt werden knnen
uint8_t IR_read(void)
{
IR_ok = 0;
return IR_in; 
}

// ----------------------------------------------------------------------------
// beendet eine bertragung nach 256 * 8us abzglich Preload
ISR(TIMER0_OVF_vect)
{
TCCR0B = T0_STOP;								// Timer0 stoppen
IR_start = 0;									// Neustart erzwingen
IR_error = 4;									// Fehlercode setzen
}

// ----------------------------------------------------------------------------
ISR(PCINT0_vect)
// wird bei Flankenwechsel an PinB.4 aufgerufen
{
static volatile uint8_t only_falling_edge = TRUE;
register uint8_t distance;

distance = TCNT0;

// wenn nur auf eine falling edge reagiert werden soll und 
// der PCINT-Pin high ist (dann liegt eine rising edge vor)
if (only_falling_edge && (PCINT_PORT & (1<<PCINT_PIN))) return;

													// Der 1. Impuls kommt vom Wakeup (falls gesendet)
													// diese fhrt aber mit 5ms Wartezeit zu Timer0_OVF und Ende.
													// und damit wieder zurck an diese Stelle
													// Beim schnelleren RC-Oscillator startet der Interrupt sofort
													// und es wird die steigende Flanke erkannt werden - aber
													// auch wieder hierhin fhren
													
													// Der 2. 1.Impuls kommt dann vom Startbit, das 1.5 Takte lang ist.
											
if (IR_start == 0)							// Fallende Flanke des Weckbits bzw. des Startbits
	{												// der Weckruf ndert nur IR_start und INitialisiert die Variablen
	IR_start = 1;								// Flag
	bitcount = 0;								// Zhler fr die eingegangenen Bits	
	IR_in    = 0;								// Variable zum Einsammeln der Bits
	IR_ok    = 0;								// wenn die 16 Bit zusammen sind, dann wird IR_ok gesetzt
	IR_error = 0;								// zeigt an, an welcher Stelle sich ein Fehler ereignet hat (timing)
	only_falling_edge = FALSE;				// falling und rising edge werden bearbeitet
	TCNT0 = 0;									// Counter zurcksetzen um die Lnge des Startbits zu beschrnken
													// fr Overflow nach 256 * 16us = 4096us
	TCCR0B = T0_START;						// Timer starten, ein Overflow setzt IR_start wieder auf '0'
	}

else if (IR_start == 1)						// die nachfolgende steigende Flanke - entweder vom Wakeup oder vom Startbit
	{												// das Startbit ist 1/2 Takt lang
	if ((distance < (hT - TO)) || (distance > (hT + TO)))	
		{											// das war dann kein Startbit
		IR_start = 0;							// neuer Start, von Vorne beginnen
		TCCR0B = T0_STOP;						// Timer abschalten
		only_falling_edge = TRUE;			// eine rising edge wird nicht weiter bearbeitet	
		}
		
	else											// das Startbit ist 1.5 Bitzeiten lang
		{											// die steigende Flanke des Startbits startet den Timer
		IR_start = 2;							// damit diese Initialisierung nur einmal am STart ausgefhrt wird
		TCNT0 = T0_PRELOAD;					// Counter zurcksetzen fr Overflow nach ca. 1.5 Bitzeit
		}
	}
	
else												// if IR_start > 1
	{
// als Prescale ist 64 eingestellt, TCCTN0 zahlt je 64 / 8.000.000 = 8us um 1 hoch
// die Bitzeit = 36/36000 = 1000us -> 1000us/8us = 125, halbe Bitzeit 62
													// wenn 1/2 Takt Abstand, dann liegt nur ein Flankenwechsel
													// zwischen zwei gleichen Bytes vor
											
	if (distance < (hT - TO + T0_PRELOAD))// zu kurz fr halben Takt
		{
		IR_start = 0;							// Neustart erzwingen
		IR_error = 1;							// Fehlerursache dokumentieren
		}
													// es handelt sich um den Flankenwechsel zwischen zwei Bits
	else if (distance < (hT + TO + T0_PRELOAD))
		{
		return;
		}
													// zu kurz fr ganzen Takt
	else if (distance < (gT - TO + T0_PRELOAD))
		{
		IR_start = 0;							// Neustart erzwingen
		IR_error = 2;							// Fehlerursache dokumentieren
		}
													// zu lang fr ganzen Takt
	else if (distance > (gT + TO + T0_PRELOAD)) 
		{
		IR_start = 0;							// Neustart erzwingen
		IR_error = 3;							// Fehlerursache dokumentieren
		}

	else											// gltiger Taktabstand ganzer Takt
		{
		TCNT0 = T0_PRELOAD;					// in Taktmitte die Bezugszeit zurcksetzen
		bitcount++;
		if (bitcount == READ_BITS) 		// Zhler fr eingegangene Bits incrementieren und testen
			{
			IR_start = 0;						// der nchste Impuls startet das Einlesen neu
			IR_ok = 1;							// Flag fr Erfolg
			}

		IR_in = IR_in << 1;					// die Daten um 1 Bit nach links schieben
		if (!(PCINT_PORT & (1<<PCINT_PIN)))// level von PCINT testen, wenn 1, dann lag eine rising edge vor
			{										// 				 			     wenn 0, dann lag eine falling edge vor
			IR_in++;								// bei einer falling edge hinten eine 1 anhngen
			}
													// Auslser steigende Flanke -> eine Null anhngen
													// das geht aber beim shiften automatisch
		}
	}
	
if (IR_start == 0)							// wenn Fehler oder Ende erreicht
	{
	TCCR0B = T0_STOP;							// Timer0 stoppen
	only_falling_edge = TRUE;				// auf eine rising edge wird nicht reagiert
													// damit wird die noch folgende rising edge bei einer "1" ignoriert
	}
}

// --- [ end of file ] --------------------------------------------------------

